@@ -11,8 +11,6 @@ use MogileFS::Server;
 use constant LIMIT => 1000;
 use constant PER_BATCH => 100;
 
-# TODO: use LWP and persistent connections to do deletes.  less local ports used.
-
 sub new {
     my ($class, $psock) = @_;
     my $self = fields::new($class);
@@ -75,6 +73,29 @@ sub work {
 
 }
 
+# deletes a given DevFID from the storage device
+# returns true on success, false on failure
+sub delete_devfid {
+    my ($self, $dfid) = @_;
+
+    # send delete request
+    error("Sending delete for " . $dfid->url) if $Mgd::DEBUG >= 2;
+
+    my $res;
+    $dfid->device->host->http("DELETE", $dfid->uri_path, undef, sub { ($res) = @_ });
+    Danga::Socket->SetPostLoopCallback(sub { !defined $res });
+    Danga::Socket->EventLoop;
+
+    my $httpcode = $res->code;
+
+    # effectively means all went well
+    return 1 if (($httpcode >= 200 && $httpcode <= 299) || $httpcode == 404);
+
+    my $status = $res->status_line;
+    error("Error: unlink failure: " . $dfid->url . ": HTTP code $status");
+    return 0;
+}
+
 sub process_tempfiles {
     my $self = shift;
     # also clean the tempfile table
@@ -212,45 +233,15 @@ sub process_deletes2 {
                 next;
             }
 
-            my $urlparts = MogileFS::Util::url_parts($path);
-
-            # hit up the server and delete it
-            # TODO: (optimization) use MogileFS->get_observed_state and don't 
-            # try to delete things known to be down/etc
-            my $sock = IO::Socket::INET->new(PeerAddr => $urlparts->[0],
-                                             PeerPort => $urlparts->[1],
-                                             Timeout => 2);
-            # this used to mark the device as down for the whole tracker.
-            # if the device is actually down, we can struggle until the
-            # monitor job figures it out... otherwise an occasional timeout
-            # due to high load will prevent delete from working at all.
-            unless ($sock) {
-                $sto->reschedule_file_to_delete2_relative($fidid,
-                    60 * 60 * (1 + $todo->{failcount}));
-                next;
-            }
-
-            # send delete request
-            error("Sending delete for $path") if $Mgd::DEBUG >= 2;
-
-            $sock->write("DELETE $urlparts->[2] HTTP/1.0\r\n\r\n");
-            my $response = <$sock>;
-            if ($response =~ m!^HTTP/\d+\.\d+\s+(\d+)!) {
-                if (($1 >= 200 && $1 <= 299) || $1 == 404) {
-                    # effectively means all went well
-                    $sto->remove_fidid_from_devid($fidid, $devid);
-                    delete $devids{$devid};
-                } else {
-                    # remote file system error?  mark node as down
-                    my $httpcode = $1;
-                    error("Error: unlink failure: $path: HTTP code $httpcode");
-
-                    $sto->reschedule_file_to_delete2_relative($fidid,
-                        60 * 30 * (1 + $todo->{failcount}));
-                    next;
-                }
+            if ($self->delete_devfid($dfid)) {
+                # effectively means all went well
+                $sto->remove_fidid_from_devid($fidid, $devid);
+                delete $devids{$devid};
             } else {
-                error("Error: unknown response line deleting $path: $response");
+                # remote file system error?  connect failure? retry in 30min
+                $sto->reschedule_file_to_delete2_relative($fidid,
+                    60 * 30 * (1 + $todo->{failcount}));
+                next;
             }
         }
 
@@ -357,37 +348,11 @@ sub process_deletes {
             next;
         }
 
-        my $urlparts = MogileFS::Util::url_parts($path);
-
-        # hit up the server and delete it
-        # TODO: (optimization) use MogileFS->get_observed_state and don't try to delete things known to be down/etc
-        my $sock = IO::Socket::INET->new(PeerAddr => $urlparts->[0],
-                                         PeerPort => $urlparts->[1],
-                                         Timeout => 2);
-        unless ($sock) {
-            # timeout or something, mark this device as down for now and move on
-            $reschedule_fid->(60 * 60 * 2, "no_sock_to_hostid");
-            next;
-        }
-
-        # send delete request
-        error("Sending delete for $path") if $Mgd::DEBUG >= 2;
-
-        $sock->write("DELETE $urlparts->[2] HTTP/1.0\r\n\r\n");
-        my $response = <$sock>;
-        if ($response =~ m!^HTTP/\d+\.\d+\s+(\d+)!) {
-            if (($1 >= 200 && $1 <= 299) || $1 == 404) {
-                # effectively means all went well
-                $done_with_devid->("deleted");
-            } else {
-                # remote file system error?  mark node as down
-                my $httpcode = $1;
-                error("Error: unlink failure: $path: HTTP code $httpcode");
-                $reschedule_fid->(60 * 30, "http_code_$httpcode");
-                next;
-            }
+        if ($self->delete_devfid($dfid)) {
+            $done_with_devid->("deleted");
         } else {
-            error("Error: unknown response line deleting $path: $response");
+            # remote file system error?  connect failure? retry in 30min
+            $reschedule_fid->(60 * 30, "http_failure");
         }
     }
 
